// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (MPU401.CC) Copyright 1998 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com




#include "bochs.h"


#if BX_USE_MPU401_SMF
bx_mpu401_c bx_mpu401;
#define this NULL
#endif

bx_mpu401_c::bx_mpu401_c(void)
{
    // nothing for now
}

bx_mpu401_c::~bx_mpu401_c(void)
{
#if (BX_MPU401_HAVELOGFILE == 1)
  fclose(LOGFILE);
#endif
#if (BX_MPU401_DATAMODE == 2)
  finishmidifile(DATAFILE);
#endif
  fclose(DATAFILE);
}

  void
bx_mpu401_c::initmidifile(FILE *midifile)
{
//  char header[] = MIDIHEADERDUMMY;
  fwrite(MIDIHEADERDUMMY, 1, MIDIHEADERSIZE, midifile);
  fwrite(MIDITRACKINIT, 1, MIDITRACKINITSIZE, midifile);
  return;
}

  void
bx_mpu401_c::finishmidifile(FILE *midifile)
{
  fpos_t tracklen;

  fwrite(MIDITRACKEND, 1, MIDITRACKENDSIZE, midifile);

  fgetpos(midifile, &tracklen);
  tracklen -= MIDIHEADERSIZE;

  fseek(midifile, MIDIHEADERSIZE - 4, SEEK_SET);

#ifdef BX_LITTLE_ENDIAN
  tracklen = (tracklen << 24) | (tracklen >> 24) |
	((tracklen & 0x00ff0000) >> 8) |
	((tracklen & 0x0000ff00) << 8);
#endif
  fwrite(&tracklen, 4, 1, midifile);

  return;
}

  void
bx_mpu401_c::init(bx_devices_c *d)
{
  BX_MPU401_THIS devices = d;

#if (BX_MPU401_HAVELOGFILE == 1)
  LOGFILE = fopen(BX_MPU401_LOGFILE,"a");	// logfile for errors etc.
#endif
  DATAFILE = fopen(BX_MPU401_DATAFILE,"w");	// the midi stream is written to this file

  /* Write midi header if datamode == 2 */
#if (BX_MPU401_DATAMODE == 2)
  initmidifile(DATAFILE);
#endif

  DATAIN init( (int) INBUFLEN);		// the data input
  DATAOUT init( (int) OUTBUFLEN);	// and data output 
  CMDPARM init( (int) CMDBUFLEN);	// and command argument
  EMULIN init( (int) EMULINBUFLEN);	// and emulator input
  EMULOUT init( (int) EMULOUTBUFLEN);	// and emulator output
  MIDICMD init( (int) MIDICMDBUFLEN);	// and midi command buffers
  BX_MPU401_THIS uartmode = 0;		// Don't start in uart mode
  BX_MPU401_THIS commandpending = 0;	// No command is pending
  BX_MPU401_THIS midipending = 0;	// No midi output pending

  writelog(4, "driver initialised, IRQ %d, IO %03x", 
	BX_MPU401_IRQ, BX_MPU401_IO);

  /* the IRQ is not implemented yet, and probably not even necessary */
//  BX_MPU401_THIS devices->register_irq(BX_MPU401_IRQ, "MPU 401");

  /* The MPU401 takes 4 IO addresses */
  for (unsigned addr=BX_MPU401_IO; addr<=BX_MPU401_IO+3; addr++) {
    BX_MPU401_THIS devices->register_io_read_handler(this,
       read_handler, addr, "MPU 401");
    BX_MPU401_THIS devices->register_io_write_handler(this,
       write_handler, addr, "MPU 401");
    }

  /* We need delta times if data mode > 0, so register a timer */
#if ( (BX_MPU401_DATAMODE > 0) && (BX_MPU401_TIMERTYPE == 1) )
  BX_MPU401_THIS timer_handle = bx_pc_system.register_timer( & BX_MPU401_THISU , midi_timer, 500000 / 384, 1, 1);
  writelog(4, "Timer initialized, handle %d", BX_MPU401_THIS timer_handle);
  BX_MPU401_THIS current_timer = 0;
#endif
}


#if (BX_MPU401_DATAMODE > 0)
#if (BX_MPU401_TIMERTYPE == 1)
void bx_mpu401_c::midi_timer (void *this_ptr)
{
  ((bx_mpu401_c *) this_ptr)->current_timer++;
  writelog(5, "TIMER event - new timer: %d", ((bx_mpu401_c *) this_ptr)->current_timer);
}
#endif

Bit32u bx_mpu401_c::gettimer(void)
{
#if (BX_MPU401_TIMERTYPE == 1)
  return (BX_MPU401_THIS current_timer);
#else
  clock_t curtime = times(NULL);
	// times -> clocks since startup, at 100 per second
	// need number of delta times, with 384*2 delta times per second:
  return ( (384 * 2 * curtime) / 100);
#endif
}
#endif


  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_mpu401_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
UNUSED(this_ptr);
#if !BX_USE_MPU401_SMF
  bx_mpu401_c *class_ptr = (bx_mpu401_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}

  Bit32u
bx_mpu401_c::read(Bit32u address, unsigned io_len)
{
#endif  // !BX_USE_MPU401_SMF

  Bit32u result = 0;

  writelog(3, "Read access to %03x for %d bytes", address, io_len);

  /* we support only byte access to the port (yet?) */
  if (io_len != 1)
    {
    writelog(1, "Read access to %03x not byte access but %d!", address, io_len);
    return(result);
    }

  if (address == STATPORT)
    {	
    result = readstatus();
    writelog(4, "Status port: result %02x", result);
    }
  else if (address == DATAPORT)
    {
    result = readdata();
    writelog(4, "Data port: result %02x", result);
    }
  else if (address == EMULPORT)
    {
    result = reademulator();
    writelog(4, "Emulator port: result %02x", result);
    }
  else
    writelog(1, "Read from unsupported port BASE+%d!", address - BX_MPU401_IO);

  /* MPU401 */
  return(result);
}


  /* Read access to the status port
   * This port is a bit coded value, with the following meaning
   * Bits 0 (LSB) to 5:	Reserved (0)
   * Bit 6: 		OUTPUT_READY - Is zero if the MPU401 accepts data on the data port
   * Bit 7 (MSB):	INPUT_AVAIL - Is zero if there is data available on the data port
   */

  Bit32u
bx_mpu401_c::readstatus(void)
{
	Bit32u result = 0;
	if (DATAIN full() != 0) result |= OUTPUT_READY;		// no output
	if (DATAOUT empty() != 0) result |= INPUT_AVAIL;	// no input
	return (result);
}

  /* Read access to data port:
   * contains an MPU_ACK after receiving a command, or
   * MPU_ACK and the command output (result).
   * This data is available in the output buffer BUFDATAOUT.
   */

  Bit32u
bx_mpu401_c::readdata(void)
{
	Bit8u res8bit;
	if ( DATAOUT get(&res8bit) == 0) {
		writelog(1, "Data port not ready - no data in buffer");
		return (PORTUNDEFINED);
	} else
	      return ( (Bit32u) res8bit);
}

  /* Read access to the emulator port:
   * This contains the emulator's responses to the various
   * commands. Data is available in EMULOUT
   */

  Bit32u
bx_mpu401_c::reademulator(void)
{
	Bit8u res8bit;
	if ( EMULOUT get(&res8bit) == 0) {
		writelog(1, "Emulator port not ready - no data in buffer");
		return (PORTUNDEFINED);
	} else
		return ( (Bit32u) res8bit);
}

  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_mpu401_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_MPU401_SMF
  bx_mpu401_c *class_ptr = (bx_mpu401_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_mpu401_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#endif  // !BX_USE_MPU401_SMF
  UNUSED(this_ptr);

  writelog(3, "Write access to %03x for %d bytes, value=%02x", address, io_len, value);

  /* we only support byte access to the port */
  if (io_len != 1) {
	writelog(1, "Write access to %03x not byte access but %d!", address, io_len);
	return;
  }

  if (address == COMDPORT) {
	writecommand(value);
	return;
  } else if (address == DATAPORT) {
	writedata(value);
	return;
  } else if (address == EMULPORT) {
	writeemulator(value);
	return;
  } else
	writelog(1, "write to unsupported port BASE+%d!", address - BX_MPU401_IO);

  return;
}
	
	/* Write access to the command port:
	 * Every byte is a command word, split into the two
	 * nibbles. More significant nibble is type of command, lower one
	 * selects the command
	 *
	 * Note: Except for Master Reset and Enter UART Mode, none of these are
	 *       supported at the moment. They will have to be implemented as
	 *       IOCTL calls to the /dev/midi00 driver
	 */

  void
bx_mpu401_c::writecommand(Bit32u value)
{
Boolean success = 0;

if (BX_MPU401_THIS commandpending != 0) {
	   /* There is a command pending. Hopefully it has all input by now...*/
	if (BX_MPU401_THIS commandbytesneeded != 0)
		writelog(1, "Incomplete command aborted, need %d more bytes!", BX_MPU401_THIS commandbytesneeded);
} else {
	   /* Every command creates an MPU_ACK response in the data port
	    * before its output */
	if (DATAOUT put(MPU_ACK) == 0)
		writelog(1, "MPU_ACK error - output buffer full");
}

switch (value >> 4) {
	case 0:	switch (value & 15) {	// MIDI commands
		case 1: writelog(4, "MIDI: Stop");
                        success=1;
			break;
		case 2: writelog(4, "MIDI: Start");
                        success=1;
			break;
		case 3: writelog(4, "MIDI: Continue");
                        success=1;
			break;
		}
		break;
	case 1:	switch (value & 15) {	// system commands
		case 5:	writelog(4, "SYST: Stop device");
                        success=1;
			break;
		}
		break;
	case 3: switch (value & 15) {	// mode commands
		case 4: writelog(4, "MODE: Return timing bytes in stop mode");
                        success=1;
			break;
		case 5: writelog(4, "MODE: Enable mode messages");
                        success=1;
			break;
		case 8: writelog(4, "MODE: Enable system common messages");
                        success=1;
			break;
		case 9: writelog(4, "MODE: Enable realtime messages");
                        success=1;
			break;
		case 15:writelog(4, "MODE: UART mode on");
			BX_MPU401_THIS uartmode=1;
			success=1;
		}
		break;
	case 8:	switch (value & 15) {	//  controller commands
		case 0:	writelog(4, "CTRL: Use internal MIDI timing");
                        success=1;
			break;
		case 1:	writelog(4, "CTRL: Use FSK timing");
                        success=1;
			break;
		case 2:	writelog(4, "CTRL: Use external MIDI timing");
                        success=1;
			break;
		case 3:	writelog(4, "CTRL: Enable metronome");
                        success=1;
			break;
		case 4:	writelog(4, "CTRL: Disable metronome");
                        success=1;
			break;
		case 7:	writelog(4, "CTRL: Enable pitch and controller");
                        success=1;
			break;
		case 10:writelog(4, "CTRL: Disable data in stop mode");
                        success=1;
			break;
		case 11:writelog(4, "CTRL: Enable data in stop mode");
                        success=1;
			break;
		}
		break;
	case 9:	switch (value & 15) {	// Clock commands
		case 4:	writelog(4, "CLCK: Disable clock to host");
                        success=1;
			break;
		case 5: writelog(4, "CLCK: Enable clock to host");
                        success=1;
                        break;
		}
		break;
	case 10:switch (value & 15) {	// Version commands
                case 12:writelog(4, "VERS: Return version number");
			if (DATAOUT put(MPU_VERSION) == 0)
				writelog(1, "Version error - output buffer full");
                        success=1;
                        break;
		case 13:writelog(4, "VERS: Return revision number");
			if (DATAOUT put(MPU_REVISION) == 0)
				writelog(1, "Revision error - output buffer full");
			success=1;
			break;
		}
		break;
        case 12:  // Timebase commands
                writelog(4, "TBSE: Set timebase");
                success=1;
                break;
        case 14:  // Tempo commands. These take one argument
		if (BX_MPU401_THIS commandpending != 0) {
			// All arguments received, execute command.
			Bit8u argument;
			CMDPARM get(&argument);
			switch (value & 15) {
        	        case 0: writelog(4, "TEMP: Set tempo %d executed", argument);
                	        success=1;
                        	break;
	                case 4: writelog(4, "TEMP: Set %d metronome clocks per click executed", argument);
        	                success=1;
                	        break;
	                case 6: writelog(4, "TEMP: Set %d metronome beats per minutes executed", argument);
        	                success=1;
                	        break;
	                case 7: writelog(4, "TEMP: Send %d clocks to host executed", argument);
        	                success=1;
                	        break;
	                }
		} else {
			BX_MPU401_THIS commandpending = value;
			BX_MPU401_THIS commandbytesneeded = 1;	// need one argument
			switch (value & 15) {
        	        case 0: writelog(4, "TEMP: Set tempo initiated");
                	        success=1;
                        	break;
	                case 4: writelog(4, "TEMP: Set metronome clocks per click initiated");
        	                success=1;
                	        break;
	                case 6: writelog(4, "TEMP: Set metronome beats per minutes initiated");
        	                success=1;
                	        break;
	                case 7: writelog(4, "TEMP: Send clocks to host initiated");
        	                success=1;
                	        break;
	                }
		}
                break;
        case 15:switch (value & 15) {   // Special commands
                case 15:writelog(4, "SPCL: Master reset of device");
			BX_MPU401_THIS uartmode=0;
                        success=1;
                        break;
                }
                break;
}
if (BX_MPU401_THIS commandpending != 0)
	writelog(4, "Command needs %d more bytes.", BX_MPU401_THIS commandbytesneeded);
if (success == 0)
	writelog(2, "Unknown command %02x, ignoring it.", value);
return;
}

	/* Write access to data port:
	 * If there is an uncomplete command pending, this data will be
	 * arguments to the command.
	 * Otherwise, this is a byte of the midi stream
	 */
  void
bx_mpu401_c::writedata(Bit32u value)
{
if (BX_MPU401_THIS commandpending != 0) {
	BX_MPU401_THIS commandbytesneeded--;
	if (CMDPARM put(value) == 0)
		writelog(1, "Command arguments too long, buffer full!");
	if (BX_MPU401_THIS commandbytesneeded <= 0) {
		   /* All argument bytes received, execute command */
		writecommand(BX_MPU401_THIS commandpending);
		BX_MPU401_THIS commandpending = 0;
	}
} else {
	writelog(3, "write to data port, value %02x", value);

	Boolean ismidicommand = 0;
	if (value >= 0x80) {
		ismidicommand = 1;
		if ( (value == 0xf7) && (MIDICMD peek(0) == 0xf0) )
			ismidicommand = 0; // it's a SysEx continue, not a command
	}

	if (ismidicommand == 1) {
		   /* This is a new command. Is the old one finished? */
		if (BX_MPU401_THIS midipending == 1) {
			writelog(4, "Midi command %d incomplete, has %d of %d bytes.", MIDICMD peek(0), MIDICMD bytes(), BX_MPU401_THIS midibytesneeded );
			writemidicommand();
			BX_MPU401_THIS midipending = 0;
		}
			/* Find out length of command in bytes */
		static unsigned eventlength[] = { 3, 3, 3, 3, 2, 2, 3, 0 };
		BX_MPU401_THIS midipending = 1;
		BX_MPU401_THIS midibytesneeded = eventlength[ ( (value & 0xf0) >> 4 ) - 8 ];

			/* Store it in the buffer (flush it first) too */
		MIDICMD flush();
		MIDICMD put(value);
	} else {
		   /* Additional bytes for old command */
		if (BX_MPU401_THIS midipending == 0) {
			writelog(1, "Midi data received, but no command pending?");
			return;
		}
		if (BX_MPU401_THIS midibytesneeded == 0) {
			BX_MPU401_THIS midibytesneeded = value;
			writelog(4, "Meta event length: %d", BX_MPU401_THIS midibytesneeded);
		} else {
			if (MIDICMD put(value) == 0) {
				writelog(1, "Midi buffer overflow!");
				BX_MPU401_THIS midibytesneeded--;
			}
			if (MIDICMD bytes() == BX_MPU401_THIS midibytesneeded) {
				writelog(4, "Midi command %d complete, has %d bytes.", MIDICMD peek(0), MIDICMD bytes() );
				writemidicommand();
				BX_MPU401_THIS midipending = 0;
			}
		}
	}	
}
}


   void
bx_mpu401_c::writeemulator(Bit32u value)
{
UNUSED(value);
	// not handled yet
}

   void
bx_mpu401_c::writedeltatime(Bit32u deltatime, FILE *midifile)
{
Bit8u outbytes[4];
int count = 0,i;

if (deltatime <= 0)
	fputc(0, DATAFILE);
else {
	while ( (deltatime > 0) && (count < 4) ) {
		outbytes[count++] = (deltatime & 0x7f) | 0x80;
		deltatime >>= 7;
	};
	outbytes[0] &= 0x7f;
	for (i=count-1;i>=0;i--)
		fputc(outbytes[i], midifile);
}
}

	/* Write the midi command contained in MIDICMD
	 * to the file where it is supposed to go
	 */

   void
bx_mpu401_c::writemidicommand(void)		
{
int i;
Bit8u value;

	/* We need to determine the time elapsed since the last command */
#if (BX_MPU401_DATAMODE != 0)
Bit32u deltatime;
static Bit32u oldtime = 0xffffffff;

if (oldtime == 0xffffffff) oldtime = BX_MPU401_THIS gettimer();

	/* Write delta time in special delta time format */
deltatime = BX_MPU401_THIS gettimer() - oldtime;
writedeltatime(deltatime, DATAFILE);

	/* Record the time */
oldtime = BX_MPU401_THIS gettimer();
#endif	// BX_MPU401_DATA != 0

i = 0;
MIDICMD get(&value);
fputc(value, DATAFILE);
if ( (value == 0xf0) || (value == 0xf7) )	// write length for sysex events
	writedeltatime(MIDICMD bytes(), DATAFILE);

while (MIDICMD empty() == 0) {
	MIDICMD get(&value);
	fputc(value, DATAFILE);
}
fflush(DATAFILE);	// to start playing immediately
}


void bx_mpu401_c::writelog(int loglevel, const char *str, ...)
{
// append a line to the log file, if desired
#if (BX_MPU401_LOGLVL > 0)
if (loglevel <= BX_MPU401_LOGLVL) {
#   if (BX_MPU401_LOGFILE != bx_printf)
	fprintf(LOGFILE, "MPU401 (%d): ", loglevel);
	va_list ap;
	va_start(ap, str);
	vfprintf(LOGFILE, str, ap);
	va_end(ap);
	fprintf(LOGFILE, "\n");
	fflush(LOGFILE);
#   else
	bx_printf("MPU401 (%d): %s", loglevel, str);
#   endif
}
#endif
}

	// the round-robin FIFO buffers of the MPU401
bx_mpu401_buffer::bx_mpu401_buffer()
{
  length = 0;		// total bytes in buffer
  head = 0;		// pointer to next slot available for new data
  tail = 0;		// pointer to next slot to be read from
  buffer = NULL;	// pointer to the actual data
}

void bx_mpu401_buffer::init(int bufferlen)
{
  if (buffer != NULL)	// Was it initialized before?
    delete buffer;

  length = bufferlen;
  buffer = new Bit8u[length];
  if (buffer == NULL)
    length = 0;		// This will be checked later
  head = 0;		// Reset the pointers
  tail = 0;
}

bx_mpu401_buffer::~bx_mpu401_buffer(void)
{
  if (buffer != NULL)
    delete buffer;
}

// Report how many bytes are available
int bx_mpu401_buffer::bytes(void)
{
  int bytes = head - tail;
  if (bytes < 0) bytes += length;
  return (bytes);
}

// This puts one byte into the buffer
Boolean bx_mpu401_buffer::put(Bit8u data)
{
  if (full() != 0)
    return 0;	// buffer full

  buffer[head++] = data;	// Write data, and increase write pointer
  head %= length;		// wrap it around so it stays inside the data

  return 1;	// put was successful
}

// This returns if the buffer is full, i.e. if a put will fail
Boolean bx_mpu401_buffer::full(void)
{
  if (length == 0)
    return 1;   // not initialized

  if ( ((head + 1) % length) == tail)
    return 1;   // buffer full

  return 0;	// buffer has some space left
}

// This reads the next available byte from the buffer
Boolean bx_mpu401_buffer::get(Bit8u *data)
{
  if (empty() != 0)
    return 0;	// buffer empty

  if (data != NULL)
	(*data) = buffer[tail];	// read data and increase read pointer
  (++tail) %= length;		// and wrap it around to stay inside the data

  return 1;	// get was successful
}

// This returns if the buffer is empty, i.e. if a get will fail
Boolean bx_mpu401_buffer::empty(void)
{
  if (length == 0)
    return 1;   // not inialized

  if (head == tail)
    return 1;   // buffer empty

  return 0;	// buffer contains data
}

// Flushes the buffer
void bx_mpu401_buffer::flush(void)
{
  tail = head;
  return;
}

// Peeks ahead in the buffer
// Warning: No checking if result is valid. Must call bytes() to check that!
Bit8u bx_mpu401_buffer::peek(int offset)
{
  return buffer[ (tail + offset) % length ];
}

