// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (MPU401.H) Copyright 1998 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com


#if BX_USE_MPU401_SMF
#  define BX_MPU401_SMF  static
#  define BX_MPU401_THIS bx_mpu401.
#  define BX_MPU401_THISU bx_mpu401
#else
#  define BX_MPU401_SMF
#  define BX_MPU401_THIS this->
#  define BX_MPU401_THISU this
#endif

#define BX_MPU401_TIMERTYPE 1	// 0: System timer, 1: Bochs timer

#if (BX_MPU401_TIMERTYPE == 0)
#  include <sys/times.h>
#endif

#define BX_MPU401_HAVELOGFILE 0
#if ( (BX_MPU401_LOGLEVEL > 0) && (BX_MPU401_LOGFILE != bx_write) )
#  define BX_MPU401_HAVELOGFILE 1
#endif

	/* Define a few shortcuts to save typing... */
#define LOGFILE BX_MPU401_THIS logfile
#define DATAFILE BX_MPU401_THIS datafile
#define DATAIN BX_MPU401_THIS bufdatain.
#define DATAOUT BX_MPU401_THIS bufdataout.
#define CMDPARM BX_MPU401_THIS bufcommandparm.
#define EMULIN BX_MPU401_THIS bufemulin.
#define EMULOUT BX_MPU401_THIS bufemulout.
#define MIDICMD BX_MPU401_THIS midicommand.

	/* The ports that we service */
#define DATAPORT	(BX_MPU401_IO)		// Port for the MIDI data (RW)
#define COMDPORT	(BX_MPU401_IO+1)	// Command port (W)
#define STATPORT	(COMDPORT)		// Status port (R)
#define CMDPPORT	(DATAPORT)		// Command parameters (W)
#define CMDRPORT	(DATAPORT)		// Command result (R)
#define EMULPORT	(BX_MPU401_IO+3)	// Emulator access port (RW)

#define OUTPUT_READY    0x40	// Bits for status port
#define INPUT_AVAIL     0x80	//
#define MPU_ACK         0xFE	// Acknowledge on data port after command
#define PORTUNDEFINED	0xFF	// Value returned if port not ready
#define MPU_VERSION	0x00	// Version number
#define MPU_REVISION	0x00	// Revision number

	/* Sizes of the internal buffers */
#define OUTBUFLEN	16	// Data port (out)
#define INBUFLEN	64	// Data port (in)
#define CMDBUFLEN	16	// Command port parameters
#define EMULINBUFLEN	16	// Emulator command port (in)
#define	EMULOUTBUFLEN	16	// Emulator command port (out)
#define	MIDICMDBUFLEN	256	// Midi commands, this large to allow for large SYSEX messages

	/* Header of the midi file. SMF type 1, 1 track, 384 deltas/quarter note */
#define MIDIHEADERDUMMY	"MThd\000\000\000\006\000\001\000\001\001\200" \
			"MTrk\177\377\377\177"
#define MIDIHEADERSIZE	22
	/* Meta event track end (0xff 0x2f 0x00) */
#define MIDITRACKEND	"\000\377\057\000"
#define MIDITRACKENDSIZE	4
	/* Set tempo to 120 (500000 us per quarter note), time sig 4/4 */
#define MIDITRACKINIT	"\000\377\121\003\007\241\140" \
			"\000\377\130\004\004\002\030\010"
#define MIDITRACKINITSIZE	15

// This is the class for the input and
// output FIFO buffers of the MPU401
class bx_mpu401_buffer {
public:

  bx_mpu401_buffer(void);
  ~bx_mpu401_buffer(void);
  void init(int bufferlen);

  Boolean put(Bit8u data);	// Write one byte in the buffer
  Boolean get(Bit8u *data);	// Read the next available byte
  Boolean full(void);		// Is the buffer full?
  Boolean empty(void);		// Is it empty?
	/* All of the above return 1 on success and 0 on error */

  void flush(void);		// Empty the buffer
  int bytes(void);		// Number of bytes in the buffer
  Bit8u peek(int offset);	// Peek ahead offset bytes, don't modify the buffer

private:
  Bit8u *buffer;
  int head,tail,length;
  };

// The actual emulator class. It intercepts accesses to the
// MPU401 IO ports and writes the raw midi stream to /dev/midi00
class bx_mpu401_c {
public:

  bx_mpu401_c(void);
  ~bx_mpu401_c(void);
  BX_MPU401_SMF void   init(bx_devices_c *);

private:

#if (BX_MPU401_HAVELOGFILE == 1)
  FILE *logfile;	// Where errors etc. are logged
#endif
  FILE *datafile;	// 
  Boolean uartmode, commandpending, midipending;
  int commandbytesneeded, midibytesneeded;
  bx_mpu401_buffer bufdataout, bufdatain, bufcommandparm;
  bx_mpu401_buffer midicommand, bufemulin, bufemulout;

#if (BX_MPU401_DATAMODE > 0)
#if (BX_MPU401_TIMERTYPE == 1)
  int timer_handle;
  int current_timer;
  static void midi_timer (void *);
#endif
  static Bit32u gettimer (void);
#endif


  bx_devices_c *devices;

  static void	writelog(int logleve, const char *str, ...);
  static void	writedata(const Bit8u value);

  static Bit32u	readdata(void);			// read access to data port
  static void	writedata(Bit32u value);	// write access to data port
  static Bit32u	readstatus(void);		// read/status port
  static void	writecommand(Bit32u value);	// write/command port
  static Bit32u	reademulator(void);		// read/emulator port
  static void	writeemulator(Bit32u value);	// write/emulator port
  static void	writemidicommand(void);		// flush the midi buffer
  static void	writedeltatime(Bit32u deltatime, FILE *midifile);
						// write in delta time coding

  static void	initmidifile(FILE *midifile);	// Write midi file header
  static void	finishmidifile(FILE *midifile);	// write track length etc.

  static Bit32u	read_handler(void *this_ptr, Bit32u address, unsigned io_len);
  static void	write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len);

#if !BX_USE_MPU401_SMF
  Bit32u	read(Bit32u address, unsigned io_len);
  void		write(Bit32u address, Bit32u value, unsigned io_len);
#endif
  };


#if BX_USE_MPU401_SMF
extern bx_mpu401_c bx_mpu401;
#endif

